from dahi import CONFIG
from dahi.bot import Bot
from dahi.context import Context
from dahi.knowledgebase import KnowledgeBase
from dahi.statement import Statement
from dahi.storages import Mongo


def getBot(botId, userId):
    print("hello")
    storage = Mongo("mongodb://localhost")
    kb = KnowledgeBase(storage, 3)
    context = Context(userId)
    bot = Bot(knowledgeBase=kb)
    print bot.respond(context, Statement("kredi"))

    print "LOGS:"
    for i in context.logs:
        print i

getBot(1, 2)


class KnowledgeBaseManager():
    def create(id):
        storage = Mongo("mongodb://localhost")
        knowledgeBase_class = import_component("knowledge_base")
        knowledgeBase_class(storage)


from dahi.document import Document
from bson import ObjectId


class Documents(object):
    def __init__(self, db):
        super(Documents, self).__init__()
        self.db = db
    
    def getAll(self):
        return (Documents.genDoc(i) for i in self.db.find())

    @staticmethod
    def genDoc(data):
        return Document(
            data["_id"],
            data["statements"],
            data["onMatch"])

    def get(self, docID):
        data = self.db.find_one({"_id": docID})
        return Documents.genDoc(data)

    def insert(self, doc):
        self.db.insert(doc.toDB())

    def update(self, doc):
        self.db.update(
            {"_id": ObjectId(doc.id)},
            doc.toDB())

    def remove(self, doc=None, docID=None):
        if doc:
            docID = doc.id
        elif not docID:
            raise AttributeError("neither document itself or its id is given")
        self.db.remove({"_id": docID})



import math

import operator

from dahi.matchers.abstracts import AbstractMatcher
from dahi.nlu import tokenize
from copy import deepcopy


class Model(object):
    """
    Term-frequency and inverse-document-frequency data model. This facilitates
    storing and querying TF and IDF values.
    """

    """
    every individual term has the following entry in the data dictionary
    """
    TERM_ENTRY = {
        "tf": {},
        "idf": 0
    }

    def __init__(self):
        super(Model, self).__init__()
        self.data = {}

    def empty(self):
        self.data = {}

    def getEntry(self, term):
        """
        returns the term entry of the model.

        :param term: term as a string
        :return: TERM_ENTRY structured data
        """
        return self.data.get(term, deepcopy(Model.TERM_ENTRY))

    def setTF(self, docId, term, frequency):
        """
        sets the term-frequency value of the given document.

        :param docId: document id
        :param term: term as a string
        :param frequency: frequency as an integer value
        :return:
        """
        entry = self.getEntry(term)
        entry["tf"][docId] = frequency
        self.data[term] = entry

    def setIDF(self, term, frequency):
        """
        sets the inverse-document-frequency of the given term

        :param term: term as a string
        :param frequency: integer
        :return:
        """
        entry = self.getEntry(term)
        entry["idf"] = frequency
        self.data[term] = entry

    def getTF(self, docId, term):
        """
        returns the term-frequency of the given document.

        if either the term or docId does not exists in the model, this will
        return 0 (zero).

        :param docId: document id
        :param term: term as a string
        :return: frequency as an integer value
        """
        # @TODO: test with not available docId and term
        entry = self.getEntry(term)
        return entry["tf"].get(docId, 0)

    def getIDF(self, term):
        """
        returns the inverse-document-frequency of the given term.

        if term does not exists in the model, this will return 0.

        :param term: term as a string
        :return: integer
        """
        entry = self.getEntry(term)
        return entry["idf"]

    def getDF(self, term):
        """
        returns the document frequency of the given term.

        document frequency is the count of how many distinct documents the term
        occurs in.

        :param term:
        :return: document frequency as an integer value
        """
        return len(self.data[term]["tf"])

    def getTerms(self):
        """
        returns all the terms found in the model.

        :return: terms list
        """
        return self.data.keys()

    def getDocIds(self, term):
        """
        returns the list of document ids in which the given term exists

        :param term: term as a string
        :return: document ids list
        """
        entry = self.getEntry(term)
        return entry["tf"].keys()

    def getScore(self, docId, term):
        """
        calculates and returns TF-IDF score according to docId and term

        :param docId: document id
        :param term: term as a string
        :return: float
        """
        tf = self.getTF(docId, term)
        idf = self.getIDF(term)
        return tf * idf


class TFIDFMatcher(AbstractMatcher):

    def __init__(self, knowledgeBase):
        super(TFIDFMatcher, self).__init__()
        self.model = self.generateModel(knowledgeBase)

    @classmethod
    def generateModel(cls, knowledgeBase):
        """
        returns a new TF-IDF model populated with the given knowledgeBase.

        :param knowledgeBase: KnowledgeBase instance
        :return: Model instance
        """
        model = Model()
        # filling the table with tf values for each term found in docs
        for doc in knowledgeBase.getAll():
            if not doc.humanSay:
                continue
            statement = doc.humanSay
            terms = tokenize(statement.text)
            n = len(terms)
            for term in terms:
                tf = model.getTF(docId=doc.id, term=term)
                tf = float(tf * n + 1) / n
                model.setTF(
                    docId=doc.id, term=term, frequency=tf)

        # filling the model with idf values for each term found in the model
        for term in model.getTerms():
            df = model.getDF(term)
            idf = math.log(knowledgeBase.count() / float(1 + df))
            if idf <= 0:
                idf = 1
            model.setIDF(term, idf)

        return model

    def getTfIdfScore(self, term, docId):
        """
        returns the TF-IDF score according to term and docId

        :param term:
        :param docId:
        :return:
        """
        tf = self.model.getTF(docId, term)
        if not tf:
            return None
        return tf * float(self.model.getIDF(term))

    def match(self, text, length=5):
        """
        matches given text against the knowledge base.

        the best matching knowledge base documents will be returned in a list.
        This list will be sorted descending by matching score.

        :param text: query text
        :param length: length of matching list to be returned
        :return: list of the best matching knowledge base documents
        """
        docScores = {}

        for term in tokenize(text):
            for docId in self.model.getDocIds(term):
                termScore = self.model.getScore(docId, term)
                docScores[docId] = docScores.get(docId, 0) + termScore

        return sorted(
            docScores.items(),
            key=operator.itemgetter(1),
            reverse=True)[:length]


if __name__ == '__main__':
    # example
    from dahi.document import Document
    from dahi.statement import Statement
    from dahi.knowledgebase import KnowledgeBase
    a = [Document(42, [Statement("elma armut elma elma")], None),
         Document(71, [Statement("karpuz armut patates")], None),
         Document(56, [Statement("elma uzum limon")], None)]

    def getCount():
        return 3

    def getAll():
        return a

    kb = KnowledgeBase(3)
    kb.getAll = getAll
    kb.count = getCount
    tt = TFIDFMatcher(kb)
    tt.generate(kb)

    print(tt.table)
    print(tt.getTfs("karp"))
    print(tt.getIdf("karp"))
    print(tt.getTfIdfScore("karp", docID=71))



from pymongo import MongoClient


def getDB():
    db = MongoClient("mongodb://localhost")["dahi"]
    return db

CONFIG = {
    "matcher": "dahi.matchers.tfidfMatcher.TFIDFMatcher",
    "nlu": "dahi.nlu.NLU",
}

#
# def import_component(name, config):
#     def load(path):
#         components = path.split('.')
#         mod = __import__(components[0])
#         for comp in components[1:]:
#             mod = getattr(mod, comp)
#         return mod
#
#     path = config[name] if config else CONFIG[name]
#     return load(path)



from dahi.document import Document
from dahi.matchers.tfidfMatcher import TFIDFMatcher
from dahi.nlu import NLU, MatchNotFound
from dahi.statement import Statement


class Bot(object):

    def __init__(self, knowledgeBase):
        self.knowledgeBase = knowledgeBase
        self.matcher = TFIDFMatcher(knowledgeBase)

    def respond(self, context, statement):
        nlu = NLU(self.matcher)

        context.humanSays(statement)

        # try:
        docID, score = nlu.findAnswer(
            statement.text, threshold=0.3, amount=4)
        doc = self.knowledgeBase.get(docID)
        if doc.botSay:
            statement = doc.botSay
        # except MatchNotFound as e:
        #     FIXME: this should not be literal, instead, knowledgeBase can be
        #     used to get this.
            # statement = Statement("sorry, I did not get it.")

        context.botSays(doc)
        return doc

    def learn(self, doc):
        self.knowledgeBase.insert(doc)



from dahi.context import Context


class ContextNotFoundError(Exception):
    def __init__(self):
        super(ContextNotFoundError, self).__init__()


class Builder(object):
    def __init__(self, storage):
        super(Builder, self).__init__()
        self.storage = storage
        self.collection = storage["contexts"]

    def get(self, contextId):
        try:
            contextRecord = self.collection.findOne({"id": contextId})
            if not contextRecord:
                raise ContextNotFoundError()
            return Context(contextId)
        except TypeError:
            raise ContextNotFoundError()

    def create(self, meta):
        contextRecord = self.collection.insert({"meta": meta})
        return self.get(contextRecord)



from dahi.document import InvalidDocument


class Context(object):
    def __init__(self, contextId):
        self.userData = {}
        self.logs = []

    def insert(self, statement):
        self.logs.append(statement)

    def botSays(self, document):
        """
        insert bot's statement into the logs

        :param document: Document instance
        :return:
        """
        assert document.botSay, InvalidDocument("botSay not found in document")
        self.insert(document.botSay)

    def humanSays(self, statement):
        self.insert(statement)



class Statement(object):
    """
    Statement is a text message container for both humans and bots.
    """

    def __init__(self, text):
        super(Statement, self).__init__()
        self.text = text

    def toJson(self):
        """

        returns instance's json representation

        :return: JSON
        """
        return {"text": self.text}

    def toDB(self):
        """
        returns instance's suitable representation for database. It actually
        converts all data into python's primitive data types.

        :return: dictionary
        """
        return self.toJson()

    def __str__(self):
        return "Statement <\"{}\">".format(self.text)

    @staticmethod
    def generate(data):
        """
        instantiate a statement object populated via data

        :param data: json
        :return: Statement
        """
        return Statement(data["text"])




from bson import ObjectId

from dahi.bot import Bot
from dahi.knowledgebase import KnowledgeBase


class BotNotFoundError(Exception):
    def __init__(self, message):
        super(BotNotFoundError, self).__init__()
        self.message = message


class Builder(object):
    def __init__(self, storage):
        super(Builder, self).__init__()
        self.storage = storage
        self.collection = storage["bots"]

    def get(self, botId):
        botRecord = self.collection.findOne({"id": botId})
        if not botRecord:
            raise BotNotFoundError("bot {} cannot be found".format(botId))

        knowledgeBase = KnowledgeBase(self.storage, botId)
        bot = Bot(knowledgeBase)
        return bot

    def create(self, botId=None, meta=None):
        if not meta:
            meta = {}

        if botId:
            botRecord = self.collection.insert({
                "_id": ObjectId(botId), "meta": meta})
        else:
            botRecord = self.collection.insert({"meta": meta})

        if botRecord:
            return self.get(botRecord)





class MatchNotFound(Exception):
    def __init__(self, msg="no answer matched", code=400):
        super(MatchNotFound, self).__init__(msg)
        self.msg = msg
        self.code = code

    def __json__(self):
        return {
            "type": self.__class__.__name__,
            "message": self.msg,
            "code": self.code}

    def __str__(self):
        return "{}: code: {}, message: {}".format(
            self.__class__.__name__,
            self.code,
            self.msg)


class NLU(object):

    def __init__(self, matcher):
        super(NLU, self).__init__()
        self.matcher = matcher

    def findBestMatch(self, matches):
        bestMatch = matches[0]
        score = bestMatch[1]
        if score > 0.3:
            return bestMatch

    def findAnswer(self, text, **kwargs):
        matches = self.matcher.match(text)

        if not matches:
            raise MatchNotFound()

        bestMatch = self.findBestMatch(matches)
        if not bestMatch:
            raise MatchNotFound()

        docID = bestMatch[0]
        score = bestMatch[1]
        return docID, score


def tokenize(text):
    return [t[:4] for t in text.split(" ")]

#
# def tf(term, document):
#     terms = tokenize(document)
#     term_frequency = terms.count(term)
#     document_size = len(terms)
#     return float(term_frequency) / document_size
#
#
# def countDocsContains(term, docs):
#     return sum(1 for i in docs if term in i)
#
#
# def idf(term, docs):
#     return math.log(len(docs) / (float(countDocsContains(term, docs))))
#
#
# def sigmoid(x):
#     return 1 / (1 + math.exp(-x))




from bson import ObjectId
from dahi.statement import Statement


class InvalidDocument(Exception):
    pass


class Document(object):
    def __init__(self, docID=None, botSay=None, humanSay=None, onMatch=None):
        super(Document, self).__init__()
        self.botSay = botSay
        self.humanSay = humanSay
        self.id = docID
        self.onMatch = onMatch

    @staticmethod
    def generate(data):
        botSay = None
        humanSay = None

        if data.get("botSay", None):
            botSay = Statement.generate(data["botSay"])

        if data.get("humanSay", None):
            humanSay = Statement.generate(data["humanSay"])

        return Document(
            docID=str(data["_id"]),
            botSay=botSay,
            humanSay=humanSay,
            onMatch=data["onMatch"])

    def __repr__(self):
        return "Document <{}>".format(self.id)

    def toJson(self):
        return {
            "_id": str(self.id),
            "botSay": self.botSay.toJson() if self.botSay else None,
            "humanSay": self.humanSay.toJson() if self.humanSay else None,
            "onMatch": self.onMatch
        }

    def toDB(self):
        return {
            "_id": ObjectId(self.id),  # FIXME: I don't like ObjectId() here
            "botSay": self.botSay.toDB() if self.botSay else None,
            "humanSay": self.humanSay.toDB() if self.humanSay else None,
            "onMatch": self.onMatch
        }



from bson import ObjectId
from dahi import getDB
from dahi.document import Document
from dahi.statement import Statement


class KnowledgeBase(object):

    def __init__(self, storageEngine, id):
        self.id = ObjectId(id)
        self.docs = []
        self.db = storageEngine["docs"]

    def getAll(self):
        return (Document.generate(i) for i in self.db.find({"kbId": self.id}))

    def get(self, docID):
        data = self.db.findOne({"kbId": self.id, "id": docID})
        return Document.generate(data)

    def insert(self, doc):
        record = doc.toDB()
        record["kbId"] = self.id
        self.db.insert(record)

    def update(self, doc):
        record = doc.toDB()
        record["kbId"] = self.id
        self.db.update(
            {"_id": ObjectId(doc.id)},
            record)

    def remove(self, doc=None, docID=None):
        if doc:
            docID = doc.id
        elif not docID:
            raise AttributeError("neither document itself or its id is given")
        self.db.remove({"kbId": self.id, "_id": docID})

    def truncate(self):
        self.db.remove({"kbId": self.id})

    def count(self):
        return self.db.count({"kbId": self.id})



from bson import ObjectId
from pymongo import MongoClient


class AbstractCollection(object):

    def find(self, pk):
        raise NotImplementedError

    def findAll(self, criteria):
        raise NotImplementedError

    def insert(self, object):
        raise NotImplementedError

    def update(self, criteria, object):
        raise NotImplementedError

    def remove(self, criteria):
        raise NotImplementedError

    def count(self, criteria=None):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError


class Collection(AbstractCollection):
    def __init__(self, coll):
        super(Collection, self).__init__()
        self.coll = coll

    @staticmethod
    def prepareCriteria(criteria):
        if criteria and "id" in criteria:
            criteria["_id"] = ObjectId(criteria["id"])
            criteria.pop("id")
        return criteria

    def findOne(self, criteria):
        criteria = Collection.prepareCriteria(criteria)
        return self.coll.find_one(criteria)

    def find(self, criteria=None):
        criteria = Collection.prepareCriteria(criteria)
        return self.coll.find(criteria)

    def insert(self, object):
        return self.coll.insert(object)

    def update(self, criteria, object):
        criteria = Collection.prepareCriteria(criteria)
        return self.coll.update(criteria, object)

    def remove(self, criteria):
        criteria = Collection.prepareCriteria(criteria)
        return self.coll.remove(criteria)

    def count(self, criteria=None):
        return self.coll.find(criteria).count()


class Mongo(object):
    def __init__(self, mongoUri):
        super(Mongo, self).__init__()
        self.db = MongoClient(mongoUri)["dahi"]

    def __getitem__(self, item):
        return Collection(self.db[item])



"""
daho
-------------

dahi

Links
`````

* `development version <http://github.com/muatik/dahi/>`

"""
import sys
from setuptools import setup

install_requires = [
]

setup(
    name='bot-dahi',
    version='0.1.11',
    url='https://github.com/muatik/dahi',
    license=open('LICENSE').read(),
    author='Mustafa Atik',
    author_email='muatik@gmail.com',
    description='question answer bot',
    keywords=[
        'bot', 'question answering'
    ],
    long_description=open('README.md').read(),
    packages=['dahi'],
    package_data={
        'dahi': [
            'matchers/*',
            ]
        },
    test_suite="tests.suite",
    zip_safe=False,
    platforms='any',
    install_requires=install_requires,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



import abc


class AbstractMatcher(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def match(self, text):
        """
        matches given text

        :param text:
        :return:
        """
        raise NotImplementedError